! string.f90 - ISO F95 cpmpatible string module for Fortran90
! Copyright (C) TOYODA Eizi, 2000.  All rights reserved.
! vi: set ts=8 sw=4:

! 
!	{W[ ISO/IEC 1539-2 : 1999 ŋK肳ĂA
!	Fortran 95 ̂߂̉ϒ񃂃W[ɑ̂
!	Fortran 90 Ŏ̂łB
!	@\B
!
!	ߎSȂƂɁAFortran 90 ɂ͎ϐ̎JȂ̂ŁA
!	\̂̂߂Ɋt͎蓮ŊJȂ
!	Ă܂B
!	̂߂̎葱Ă܂ ISO ƌ݊ł
!	ȂȂ邽߁AȂ񂾂ςł͂邯ǁA{Cuł͒[
!	Zbg邱ƂpƂ݂ȂƂɂB

module iso_varying_string

    implicit none

    private
    public varying_string, len, var_str, char, get, put, put_line, &
	& assignment(=), operator(==), operator(/=), operator(//), &
	& operator(<), operator(<=), operator(>), operator(>=), &
	& table_entry

    type varying_string
    	private
	integer::				id
    end type

    ! Iɕێ镶\̌^
    type table_entry
	! nlink: ɂ 0 ɂȂB
	! l𐶐Ƃɂ͑ȂB
	! VARYING_STRING ϐɊi[Ƃ 1 A
	! ϐsvɂȂƂ 1 B
	! 1  0 ɂȂƂ buffer pׂłB
    	integer::			nlink
	! ɂċԂɂȂ
	character, pointer::		buffer(:)
    end type

    interface len
	module procedure string_len
    end interface

    interface var_str
	module procedure char_to_string
    end interface

    interface char
	module procedure string_to_char_all
	module procedure string_to_char_length
    end interface

    interface get
	module procedure string_get
	module procedure string_get_default
    end interface

    interface put
	module procedure string_put
	module procedure string_put_default
	module procedure char_put
	module procedure char_put_default
    end interface

    interface put_line
	module procedure string_put_line
	module procedure string_put_line_default
	module procedure char_put_line
	module procedure char_put_line_default
    end interface

    interface scan
	module procedure string_scan_string
	module procedure string_scan_char
    end interface

    interface verify
	module procedure string_verify_string
	module procedure string_verify_char
    end interface

    interface index
	module procedure string_index_string
	module procedure string_index_char
	module procedure char_index_string
    end interface

    interface assignment(=)
	module procedure string_let_char
	module procedure char_let_string
	module procedure string_let_string
    end interface

    interface operator(//)
	module procedure string_add_string
	module procedure char_add_string
	module procedure string_add_char
    end interface

    interface operator(==)
	module procedure string_eq_string
	module procedure string_eq_char
	module procedure char_eq_string
    end interface

    interface operator(/=)
	module procedure string_ne_string
	module procedure string_ne_char
	module procedure char_ne_string
    end interface

    interface operator(<)
	module procedure string_lt_string
	module procedure string_lt_char
	module procedure char_lt_string
    end interface

    interface operator(<=)
	module procedure string_le_string
	module procedure string_le_char
	module procedure char_le_string
    end interface

    interface operator(>)
	module procedure string_gt_string
	module procedure string_gt_char
	module procedure char_gt_string
    end interface

    interface operator(>=)
	module procedure string_ge_string
	module procedure string_ge_char
	module procedure char_ge_string
    end interface

    !
    ! rẐ߂̃V{
    !

    integer, parameter::	CMP_LESSTHAN = -1
    integer, parameter::	CMP_EQUIVALENT = 0
    integer, parameter::	CMP_GREATERTHAN = 1
    integer, parameter::	CMP_ERROR = -100

    ! --- \ ---
    !
    !  table_entry \̂Ȃ Table Ɋi[B
    ! Table  ɏ߂ĊmۂƂɂ͑傫 INITIAL_TABLE_SIZE
    ! ł邪Ao^̈g傷Kv邽т2{ɊgB

    integer, parameter:: 	INITIAL_TABLE_SIZE = 4

    type(table_entry), save, pointer::	Table(:)

contains

    !
    ! === J葱 ===
    !

    !
    !  len ֐̎
    !

    integer function string_len(string) result(result)
	type(varying_string), intent(in)::	string
    continue
    	result = size(Table(string%id)%buffer)
    end function

    !
    !  var_str ֐̎
    !

    type(varying_string) function char_to_string(char) result(result)
	character(len=*), intent(in)::	char
    continue
    	!  result = char ̓NJE^1̂Ŗ]܂Ȃ
	result%id = table_set_char(char)
    end function

    !
    !  char ֐̎
    !

    function string_to_char_all(string) result(result)
	type(varying_string), intent(in)::	string
	character(len = size(Table(string%id)%buffer))::	result
    continue
	call char_let_string(result, string)
    end function

    function string_to_char_length(string, length) result(result)
	type(varying_string), intent(in)::	string
	integer, intent(in)::			length
	character(len = length)::		result
    continue
	call char_let_string(result, string)
    end function

    !
    ! ̑̎
    !

    subroutine string_let_string(lhs, rhs)
	type(varying_string), intent(out)::	lhs
	type(varying_string), intent(in)::	rhs
    continue
	call string_link_decrement(lhs%id)
	lhs%id = rhs%id
	call table_link_increment(rhs%id)
    end subroutine

    subroutine string_let_char(string, char)
	type(varying_string), intent(out)::	string
	character(len=*), intent(in)::		char
    continue
	call string_link_decrement(string%id)
	string%id = table_set_char(char)
	call table_link_increment(string%id)
    end subroutine

    subroutine char_let_string(char, string)
	character(len=*), intent(out)::		char
	type(varying_string), intent(in)::	string
    continue
	char = table_get_char(string%id)
    end subroutine

    !
    !  // Zq̎
    !

    function string_add_string(lhs, rhs) result(result)
	type(varying_string)::			result
	type(varying_string), intent(in)::	lhs, rhs
	integer::				lhslen, rhslen
    continue
write(0, "('#string_add_string:', 2i4)"), lhs%id, rhs%id
	lhslen = len(lhs)
	rhslen = len(rhs)
	if (lhslen == 0) then
	    result%id = rhs%id
	    return
	else if (rhslen == 0) then
	    result%id = lhs%id
	    return
	endif
	result%id = table_allocate(lhslen + rhslen)
	if (result%id <= 0) return
	Table(result%id)%buffer(1: lhslen) = Table(lhs%id)%buffer(1: lhslen)
	Table(result%id)%buffer(lhslen+1: lhslen+rhslen) = &
	    & Table(rhs%id)%buffer(1: rhslen)
write(0, "('#string_add_string:', 3i4)"), result%id, lhs%id, rhs%id
write(0, *) '#<', Table(result%id)%buffer, '>=<', Table(lhs%id)%buffer(:), &
& '>+<', Table(rhs%id)%buffer(:), '>'
    end function

    function char_add_string(char, string) result(result)
	type(varying_string)::			result
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_add_string(var_str(char), string)
    end function

    function string_add_char(string, char) result(result)
	type(varying_string)::			result
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_add_string(string, var_str(char))
    end function

    !
    ! ̉Zq == ̎
    !

    logical function string_eq_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_EQUIVALENT
    end function

    logical function string_eq_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_eq_string(string, var_str(char))
    end function

    logical function char_eq_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_eq_string(var_str(char), string)
    end function

    !
    ! ̉Zq /= ̎
    !

    logical function string_ne_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = .not. string_eq_string(lhs, rhs)
    end function

    logical function string_ne_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = .not. string_eq_char(string, char)
    end function

    logical function char_ne_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = .not. char_eq_string(char, string)
    end function

    !
    ! ̉Zq < ̎
    !

    logical function string_lt_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_LESSTHAN
    end function

    logical function string_lt_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_lt_string(string, var_str(char))
    end function

    logical function char_lt_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_lt_string(var_str(char), string)
    end function

    !
    ! ̉Zq <= ̎
    !

    logical function string_le_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_LESSTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_le_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_le_string(string, var_str(char))
    end function

    logical function char_le_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_le_string(var_str(char), string)
    end function

    !
    ! ̉Zq > ̎
    !

    logical function string_gt_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_GREATERTHAN
    end function

    logical function string_gt_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_gt_string(string, var_str(char))
    end function

    logical function char_gt_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_gt_string(var_str(char), string)
    end function

    !
    ! ̉Zq >= ̎
    !

    logical function string_ge_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_GREATERTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_ge_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_ge_string(string, var_str(char))
    end function

    logical function char_ge_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_ge_string(var_str(char), string)
    end function

    !
    ! o
    !

    subroutine string_get_default(string, maxlen, iostat)
	type(varying_string), intent(out)::	string
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
    continue
	call string_get(-1, string, maxlen, iostat)
    end subroutine

    subroutine string_get(unit, string, maxlen, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(out)::	string
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
	integer::		alreadyread, buflen, nowread, ios, maxsize
	integer, parameter::	BUFFERSIZE = 80
	character(len = BUFFERSIZE)::		buffer
    continue
	if (present(maxlen)) then
	    maxsize = maxlen
	else
	    maxsize = huge(maxsize)
	endif
	alreadyread = 0
	string = ''
	do
	    if (alreadyread >= maxsize) return
	    buflen = min(BUFFERSIZE, maxsize - alreadyread)
	    if (unit >= 0) then
		read(unit=unit, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    else
		read(unit=*, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    endif
	    if (ios /= 0) then
		if (present(iostat)) then
		    iostat = ios
		    return
		else
		    print *, 'get_string: read error ', ios
		    stop
		endif
	    endif
	    alreadyread = alreadyread + nowread
	    string = string // buffer(1: nowread)
	enddo
	if (present(iostat)) iostat = 0
	return
	! in case of EOR
100	string = string // buffer(1: nowread)
	if (present(iostat)) iostat = 0
    end subroutine string_get

    subroutine char_put_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put(-1, char, iostat)
    end subroutine

    subroutine char_put(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'char_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_default(string, iostat)
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
    continue
	call string_put(-1, string, iostat)
    end subroutine

    subroutine string_put(unit, string, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char(string)
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char(string)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'string_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine char_put_line_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put_line(-1, char, iostat)
    end subroutine

    subroutine char_put_line(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'char_put_line: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_line_default(string, iostat)
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
    continue
	call string_put_line(-1, string, iostat)
    end subroutine

    subroutine string_put_line(unit, string, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char(string)
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char(string)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'string_put_line: write error ', ios
	    endif
	endif
    end subroutine

    !
    ! --- gݍ݊֐̏㏑ ---
    !

    !
    ! index ̑p
    !

    function string_index_string(string, substring, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, substring
	logical, optional::			back
    continue
	result = index(char(string), char(substring), back)
    end function

    function string_index_char(string, substring, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(char(string), substring, back)
    end function

    function char_index_string(string, substring, back) result(result)
	integer::				result
	character(len = *), intent(in)::	string
	type(varying_string), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(string, char(substring), back)
    end function

    !
    ! scan ̑p
    !

    function string_scan_string(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, set
	logical, optional::			back
    continue
	result = scan(char(string), char(set), back)
    end function

    function string_scan_char(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = scan(char(string), set, back)
    end function

    !
    ! verify ̑p
    !

    function string_verify_string(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, set
	logical, optional::			back
    continue
	result = verify(char(string), char(set), back)
    end function

    function string_verify_char(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = verify(char(string), set, back)
    end function

    !
    ! === Iɗp ===
    !

    ! rZq
    integer function string_cmp_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	type(table_entry), pointer::		lhs_entry, rhs_entry
	integer::				lhs_len, rhs_len, i
	character::				lhs_char, rhs_char
    continue
	lhs_entry => Table(lhs%id)
	rhs_entry => Table(rhs%id)
	if (.not. associated(lhs_entry%buffer) &
	& .or. .not. associated(rhs_entry%buffer)) then
	    result = CMP_ERROR
	    return
	endif
	result = CMP_EQUIVALENT
	lhs_len = len(lhs)
	rhs_len = len(rhs)
	do, i = 1, max(lhs_len, rhs_len)
	    if (i <= lhs_len) then
		lhs_char = lhs_entry%buffer(i)
	    else
		lhs_char = ' '
	    endif
	    if (i <= rhs_len) then
		rhs_char = rhs_entry%buffer(i)
	    else
		rhs_char = ' '
	    endif
	    if (lhs_char > rhs_char) then
		result = CMP_GREATERTHAN
		return
	    else if (lhs_char < rhs_char) then
		result = CMP_LESSTHAN
		return
	    endif
	enddo
    end function

    !
    ! --- \̑ ---
    !

    ! \̏
    ! ݂̏ꍇQɏI
    function initTable()
	logical::		initTable
	integer::		status, i
    continue
	initTable = .true.
	if (associated(Table)) return
	allocate(Table(INITIAL_TABLE_SIZE), stat=status)
	if (status /= 0) then
	    initTable = .false.
	    return
	endif
	! [̕p
	Table(1)%nlink = 0
	allocate(Table(1)%buffer(1:0))
	! ȊȌ
	do, i = 2, INITIAL_TABLE_SIZE
	    Table(i)%nlink = -1
	    nullify(Table(i)%buffer)
	enddo
    end function

    ! \̍ŏ̎gĂȂT
    integer function firstEmptyId()
	integer::			i
    continue
	do, i = 1, size(Table)
	    if (Table(i)%nlink < 0) then
		firstEmptyId = i
		return
	    endif
	enddo
	firstEmptyId = -1
    end function

    ! \̑傫ς
    ! (sĂ񍐂Ȃ)
    subroutine growTable(newsize)
	integer, intent(in)::		newsize
	integer::			copysize, status, i
	type(table_entry), pointer::	newtable(:)
    continue
	copysize = min(size(Table), newsize)
	allocate(newtable(newsize), stat=status)
	if (status == 0) then
	    newtable(1:copysize) = Table(1:copysize)
	    newtable(copysize + 1: newsize)%nlink = -1
	    do, i = copysize + 1, newsize
		nullify(newtable(i)%buffer)
	    enddo
	    deallocate(Table)
	    Table => newtable
	endif
    end subroutine

    ! \ id Ԗڂ̎̕QƃJE^1
    subroutine table_link_increment(id)
    	integer, intent(in)::		id
    continue
	if (id <= 1 .or. id > size(Table)) return
	if (Table(id)%nlink < 0) return
write(0, "('#table_link_increment', 2i4)") id, Table(id)%nlink
	Table(id)%nlink = Table(id)%nlink + 1
    end subroutine

    ! \ id Ԗڂ̎̕QƃJE^1
    subroutine string_link_decrement(id)
	integer, intent(in)::		id
    continue
	if (id <= 1 .or. id > size(Table)) return
	if (Table(id)%nlink < 0) return
write(0, "('#table_link_decrement', 2i4)") id, Table(id)%nlink
	Table(id)%nlink = Table(id)%nlink - 1
	if (Table(id)%nlink == -1) then
	    print "(a, i4, a)", 'string_link_decrement: link counter #', id, &
		& ' is decremented but not incremened'
	else if (Table(id)%nlink == 0) then
	    deallocate(Table(id)%buffer)
	    nullify(Table(id)%buffer)
	    Table(id)%nlink = -1
	endif
    end subroutine

    ! \ id Ԗڂ̕Ԃ
    function table_get_char(id) result(result)
	integer, intent(in)::		id
	character(len = size(Table(id)%buffer))::	result
	integer::			length 
    continue
    	length = size(Table(id)%buffer)
	result(1: length) = &
	    & transfer(Table(id)%buffer(1: length), result(1: length))
    end function

    ! id Ԗڂ̈̂̕߂̃p
    logical function table_dispose(id) result(result)
	integer, intent(in)::			id
	type(table_entry), pointer::		cur
    continue
	if (id <= 0 .or. id > size(Table)) then
	    result = .FALSE.
	    return
	endif
	result = .TRUE.
write(0, "('#table_dispose', i4)") id
	cur => Table(id)
	if (.not. associated(cur%buffer)) return
	deallocate(cur%buffer)
	nullify(cur%buffer)
	cur%nlink = -1
    end function

    ! V̈mۂ length ̗̈tB
    ! ̈ԍ (s畉̒l) Ԃ
    integer function table_allocate(length) result(result)
	integer, intent(in)::			length
	integer::				status
	type(table_entry), pointer::		cur
    continue
	! `FbNɎs畉̒lԂ
	result = -1
	! `FbN
	if (.not. initTable()) return
	! ̒`FbN
	if (length < 0) return

	! ̒͂ЂƂvȂ
	if (length == 0) then
	    result = 1
write(0, "('#table_allocate', 2i4)") length, result
	    return
	endif

	! ̒ȂΕ̈m
	result = firstEmptyId()
	if (result < 0) then
	    call growTable(size(Table) * 2)
	    result = firstEmptyId()
	    if (result < 0) return
	endif
	cur => Table(result)
	if (associated(cur%buffer)) deallocate(cur%buffer)
	allocate(cur%buffer(1:length), stat=status)
	if (status /= 0) result = -3
write(0, "('#table_allocate', 2i4)") length, result
    end function

    ! \ɕ char o^ԍԂ
    integer function table_set_char(char) result(result)
	character(len=*), intent(in)::	char
	integer::			length
    continue
	length = len(char)
	result = table_allocate(length)
	if (result < 0) return
	Table(result)%buffer = transfer(char(1: length), Table(result)%buffer)
	Table(result)%nlink = 0
write(0, "('#table_set_char', 2i4, '<', a, '>')") length, result, char
    end function

end module iso_varying_string
